/* iomain.c
/* Copyright (c) 1988, Prime Computer, Inc., Natick, MA 01760
/*                     All Rights Reserved
*/

/* This is driver program for IOBENCH test.  It creates and populates the
   necessary files, launches the daemon processes that do the benchmark
   proper, and then collects, summarizes, and reports the benchmark
   results.  Although IOBENCH is meant to be operating system independent,
   it does things in the UNIX style.  Unless absolutely necessary, non
   portable UNIX features are not used. 

   The overall plan for the benchmark is to create NF files each containing
   NR RECLEN-byte records.  The records have implicit keys, which are longs
   in the range 0...NR-1.  The only interesting part of each record is a
   long counter.  NU daemon users are created.  Each user does
   "transactions" consisting of R reads and W writes with an optional CP
   load.  IOBENCH can be run in "timed" mode, in which all of the users do
   as many transactions as they can in the specified time interval, or in
   "counted" mode, in which case each user does its share (1/NU) of the
   specified number of transactions as quickly as it can.  In timed mode,
   we're interested in how many transactions were done; in counted mode,
   we're interested in how long each user took.  See "iouser.c" for details
   on transactions. 

   Each of the user daemons writes a result record to record K of the
   results file containing statistics such as number, of transactions
   completed, CP time used, etc.  Detailed statistics of system activity
   are gathered by phantom NU+1, which runs PRIMOS usage, UNIX sar, etc. 
   and writes its results into a file. 

   Each of the user processes "locks" the records it reads and writes
   during the period in which it is using them.  This locking is a
   simulation of what would be done in a "real" data management system; the
   only requirement is that no other user be able to access the locked
   record until locking user releases the lock.  IOBENCH is distributed
   with locking code implemented for some architectures in terms of atomic
   instructions such as "test and set"; this code is activated by compiling
   IOBENCH with REALLOCKING defined.  If you're porting IOBENCH to a new
   architecture, you'll have to implement about half a page of assembly
   code.  You can build IOBENCH with REALLOCKING not defined, in which case
   "locking" is implemented via ordinary non-atomic C code; this
   approximates the overhead that would be incurred from real locking.  See
   "iolocks.c" for more information on locking. 

   IOBENCH is controlled by an input file that contains the following
   information (see "ioparams.c" for more details):

         runid
         number of files
         number of records (or total file length)
         record size
         number of users
         locks per file
         lock sleep time
         start delay
         run duration
         end delay
         cpu work parameter
         reads per transaction
         writes per transaction (must be <= reads per transaction)
         pathname for file 1      OR   @filespath
         pathname for file 2
         ...

   The file pathnames are provided in the control file so that IO load can
   be spread across multiple disks and controllers.  If "@filespath" is
   specified, file names are read from file with specified pathname instead
   of from original control file. 

   Synchronization and control of the user processes is done by an ad hoc
   scheme that uses the file system and avoids operating system
   dependencies. 

   Startup of the background processes is controlled by a sync file.  Each
   background task does its initialization and then waits until the
   controlling process creates the sync file.  As soon as the sync file
   appears, the background processes start their timed period. 

   As soon as each background process starts running, it creates a unique
   run file using the name passed to it by the controlling process.  The
   background processes delete their unique run files as the last thing
   they do.  Thus, the existance of any run file means the process
   associated with that task is still active and the date in the results
   file is not yet valid. 

   IOBENCH is a "relative", not an "absolute" benchmark.  The most useful
   way to use IOBENCH is to define a spectrum of runs that vary key
   parameters such as number of users, CP work component, or record length
   over a range of 10-20 values that cover both IO limited and CP limited
   performance.  Do the set of runs on a base configuration, and then redo
   them on a different configuration (e.g.  different hardware platform,
   different OS, different IO controller, etc) and compare the results
   against the base configuration. 

   The name of the input control file is used to generate the names of two
   output files.  control.res contains the results of the run and
   control.use contains system usage data for the run.  An optional summary
   file can be created containing one line for each test configuration.

   Initial Version: 10 August 1988 by Barry Wolman

          Modified: 16 August 1988 by Barry Wolman
              1. extend file lengths
              2. run iostat background process
              3. query environment

          Modified: 20 August 1988 by Barry Wolman
              1. Backfit results of port to UNIX
              2. Add version ID to output

          Modified: 28 August 1988 by Barry Wolman
              1. Deal with case where a background process doesn't terminate
              2. Speed up termination check
              3. Store number of users in results file

          Modified: 29 August 1988 by Barry Wolman
              1. Print process ids of background tasks
              2. Kill processes if error
              3. Add work parameter
              4. Fixed bug in file population (needed seek to EOF) when making
                 existing file larger

          Modified: 6 September 1988 by Barry Wolman
              1. Print volume of data processed
              2. Identify run file in release message

          Modified: 10 September 1988 by Barry Wolman
              1. Close output file before spawning UNIX daemons
              2. Reopen output file after UNIX daemons finish
              3. Summarize file closing and opening due to file unit limits
              4. Show run duration instead of finish time
              5. Add debug flag to control printing of process ids
              6. Use SIGTERM instead of SIGKILL for terminating user procs
              7. Catch SIGTERM & SIGINT (UNIX) and QUIT$ PRIMOS and do a
                 cleanup

          Modified: 16 September 1988 by Barry Wolman
              1. Calculate adjusted totals if some runs didn't finish
              2. Kill phantoms if users quits (PRIMOS)
              3. Sleep 10 seconds and LOOP 10 times when checkin termination
              4. Print run summary in grep-able form

          Modified: 26 September 1988 by Barry Wolman
              1. Make PRIMOS Phantom logout messages come out AFTER all of
                 our output.
              2. Eliminate phantom logout message if not debug (PRIMOS)

          Modified: 27 September 1988 by Barry Wolman
              1. Set RW lock on results file (PRIMOS)
              2. Print statistics gathered by phantoms

          Modified:  2 October 1988 by Barry Wolman
              1. Skip IO if calibrate flag set
              2. Sleep 30 seconds when checking termination
              3. Rewrote termination check to sleep AFTER check not BEFORE

          Modified: 15 February 1989 by Tom Olson
              1. Added transaction grouping,
              2. Added read/write ratio (n_reads and n_writes) parameters
              3. Compute transaction response time distribution statistics.

          Modified: 27 February 1989 by Barry Wolman
              1. Call init_params to initialize params from defaults and
                 environment
              2. Parse keyword style contro lfile
              3. Create interim control file for user processes
              4. Write line to summary file
              5. Move in check_params and support subroutines from iosubs
              6. Minor cleanup

         Modified: 28 March 1989 by Barry Wolman
              1. Correct calculation of total MB (Olson's Bug!)
              2. Show TPS summary even if -t not specified, -t now controls
                 only display of detailed histogram
              3. Allow input file to be empty

         Modified:  3 April 1989 by Barry Wolman
              1. Make input file optional
              2. Add -a, -h, and -u options
              3. Use UPPERCASE for temp file names
              4. Delete RESULTS file at end of run now that prt_results
                 is defunct.

         Modified: 25 May 1989 by Barry Wolman
              1. Add variant capability

         Modified: 24 August 1989 by Barry Wolman
              1. Don't start server if variant is 0
              2. Don't compile server code if no message queue facility

         Modified:  2 May 1990 by Barry Wolman
              1. Keep going with reduced number if users if all user
                 processes can't be started

         Modified: 18 May 1990 by Barry Wolman
              1. LOCKS=0 omits locking
              2. Pass delay parameter to usage process
              3. Show number of lock hits

         Modified:  6 June 1990 by Barry Wolman
              1. Adapt to SPEC Makefile conventions

         Modified: 20 June 1990 by Barry Wolman
              1. Keep user & sys times separate, report total times
              2. Move parameter processing to ioparams.c
              3. Use filelen instead of total_bytes
              4. All durations in seconds
              5. Call ioflush() to force out cached pages
              6. Loop over output files
*/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "ioincl.h"
#include "iodefaults.h"
#include "ioserver.h"

#include <signal.h>
#include <fcntl.h>
int cleanup();

#define SYNCFILE   "IOSYNCFILE"
#define RUNFILE    "IORUN"
#define RESULTS    "IORESULTS"
#define CTLFILE    "IOCONTROLFILE"
#define OUTFILE    "iobench"

#define MAXLOOPS   10

/* Following macro is used to iterate over the set of output files */

#define OFLOOP for(z=0; of[z]; z++)

int unit[MAX_FILES];

struct control control;

long start_time;

FILE *of[4], *fopen();

extern char *version,*flags,*compdate;
extern struct keyword_table keyword[];

char *getenv(),*copystr();
long ioflush();

int debug = 0, show_tps = 0, append = 0, help = 0, usage = 0;
int show_checkpoints = 0;
char *iosummary = NULL;
char *iooutput = NULL;

int pid[MAX_USERS],finished[MAX_USERS];
int spawned_users = 0;
int smsgid,rmsgid,mypid,serverpid,user;
struct msgstruct smsg,rmsg;
char *instrs[] = {
    "Usage: iomain [options] [control_file]",
    "",
    "Options are: -a         append to output file",
    "             -c         show checkpoints",
    "             -d         set debug flag",
    "             -h         help, print this message",
    "             -ofile     use 'file' as prefix of output files",
    "             -sfile     write summary line to 'file'",
    "             -t         show TPS statistics",
    "             -u         collect system usage statistics",
    "             -vN        set variant to N, where N is octal #",
    "",
    "Control file can be omitted if all parameters are specified via",
    "environment or defaults.",
    "",
    "Contact Barry Wolman (barry@s66.prime.com) if you have questions.\n",
    NULL};

main(argc,argv)
char **argv;
    {
    register int i,j,n;
    int loops,rfid,errs,hits,z;
    int nc,no,total_nc,total_no,opens[MAX_FILES],closes[MAX_FILES];
    int uses[MAX_FILES],worst_rec[MAX_FILES],worst_user[MAX_FILES];
    long reads[MAX_FILES],writes[MAX_FILES];
    double sum_time[MAX_FILES],sumsq_time[MAX_FILES],worst_time[MAX_FILES];
    int failed,completed,avg_n;
    double usert1,usert2,syst1,syst2,mb,avg_usert,avg_syst,avg_time,stddev_time;
    double variance_time,samples;
    double server_usert,server_syst,ut,st,tt,flush_usert,flush_syst;
    char *p,*userargv[10],temp[16],runfile[32],outname[256],usename[256];
    struct rec rec;
    struct results results;
    struct checkpoint checkpoint[NUMBER_OF_CHECKPOINTS+1],*cp;
    double avg_resp,total_avg_resp,tps;
    long et,maxet,ft,totaltime,cum_resp_dist[26];
    int k,old;
    struct msgtime *tp;

    /* Check for SUMMARY and OUTPUT specified in environment so that
       specification on command line can override these */

    if (p = getenv("SUMMARY")) iosummary = copystr(p);
    if (p = getenv("OUTPUT")) iooutput = copystr(p);

    /* Now process arguments and options */

    errs = 0;
    control.variant = 0;
    while(argc > 1 && argv[1][0] == '-') {
         switch(argv[1][1]) {

         case 'a':
         case 'A':
                   append++;
                   break;

         case 'c':
         case 'C':
                   show_checkpoints++;
                   break;

         case 'd':
         case 'D':
                   debug++;
                   break;

         case 'h':
         case 'H':
                   for(i = 0; instrs[i]; i++) fprintf(stderr,"%s\n",instrs[i]);
                   help++;
                   break;

         case 'o':
         case 'O':
                   iooutput = argv[1][2] ? argv[1]+2 : NULL;
                   break;

         case 's':
         case 'S':
                   iosummary = argv[1][2] ? argv[1]+2 : NULL;
                   break;

         case 't':
         case 'T':
                   show_tps++;
                   break;

         case 'u':
         case 'U':
                   usage++;
                   break;

         default:  fprintf(stderr,"iobench: '%c' is not a valid option\n",argv[1][1]);
                   errs++;
                   }

         argv++;
         argc--;
         }

    if (help || errs) return;

    /* If no output file specified via environment, use name of input file
       if present, otherwise use default */

    if (iooutput == NULL) iooutput = (argc > 1 ? argv[1] : OUTFILE);

    sprintf(outname,"%s.res",iooutput);
    sprintf(usename,"%s.use",iooutput);

    if (append) {
         of[1] = fopen(outname,"a");
         }
    else of[1] = fopen(outname,"w");

    if (of[1] == NULL) {
         fprintf(stderr,"iobench: Could not (re)open %s\n",outname);
         return;
         }

    of[0] = stdout;
    of[2] = of[3] = NULL;

    time(&start_time);

    init_params();
    if (argc > 1 && parse_control(argv[1])) return;

    if (check_params()) return;

    /* Either LOCKS=0 (new style) or SLEEP<0 (old style) surpresses locking */

    if (control.lock_sleep_time < 0) control.locks_per_file = 0;

    /* We can't assume that user processes will have access to environment
       variables, so we'll create a complete copy of control file */

    if (dump_control(CTLFILE)) {
         fprintf(stderr,"iobench: Could not copy control file\n");
         return;
         }

    OFLOOP print_control(of[z]);

    /* Look for optional environment variables that define system
       configuration */

    if (p = getenv("SYSTEM")) fprintf(of[1],"System is %s\n",p);
    if (p = getenv("MEMORY")) fprintf(of[1],"Memory is %s\n",p);
    if (p = getenv("CONTROLLERS")) fprintf(of[1],"Controllers is %s\n",p);
    if (p = getenv("BUFFERS")) fprintf(of[1],"Buffers is %s\n",p);
    if (p = getenv("COMMENT")) fprintf(of[1],"Comment is %s\n",p);

    /* Calibrate CPU work for each transaction */

    if (control.param) {
         cptimes(&usert1,&syst1);
         work(control.param);
         cptimes(&usert2,&syst2);
         results.work_cp = usert2-usert1;
         }
    else results.work_cp = 0.0e0;

    OFLOOP fprintf(of[z],"CP load/transaction = %.3f secs\n",results.work_cp);

    rec.counter = 1;
    for(i = 0; i < control.reclen; i++) rec.junk[i] = 0;

    for(i = 0; i < control.n_files; i++) {
         opens[i] = closes[i] = uses[i] = reads[i] = writes[i] = 0;
         sum_time[i] = sumsq_time[i] = 0.0e0;
         worst_rec[i] = worst_user[i] = -1;
         worst_time[i] = -1.0e0;

         if ((unit[i] = open(control.path[i],2)) < 0) {
              if ((unit[i] = creat(control.path[i],0666)) < 0) {
                   fprintf(stderr,"iobench: Could not open/create %s\n",control.path[i]);
                   return;
                   }
              }

         if(populate(i,&rec)) return;;
         }

    printf("Finished file population\n");

    /* Users = 0 means just file population */

    if (control.n_users == 0) return;

    /* Initialize the lock manager */

    initlocks(control.n_users,control.n_files,control.locks_per_file,
     control.variant & LOCKING_VARIANT);

    /* Ensure proper cleanup */

    signal(SIGTERM,cleanup);
    signal(SIGINT,cleanup);

    /* Delete the sync file used to control start up of users */

    unlink(SYNCFILE);

    /* Create the results file */

    unlink(RESULTS);
    if ((rfid = creat(RESULTS,0666)) < 0) {
         fprintf(stderr,"iobench: Could not create %s\n",RESULTS);
         return;
         }

    /* Initialize the RESULTS contents so we can ensure that background
       processes actually wrote their results properly */

    results.counter = -1;
    results.start = results.finish = 0;
    results.usert = results.syst = 0.0e0;
    results.n_files = control.n_files;
    results.n_recs = control.n_recs;
    results.n_users = control.n_users;
    results.locks_per_file = control.locks_per_file;
    results.reclen = control.reclen;
    results.lock_sleep_time = control.lock_sleep_time;
    results.duration = control.duration;
    results.param = control.param;

    for (i = 0; i <= 25; i++) results.resp_dist[i]=0;

    write(rfid,&results,sizeof(results));   /* for user 0 */

    for (i= 0; i < 26; i++) cum_resp_dist[i]=0;

    /* Each background process will create a RUNFILEnn when it starts
       running and delete that file as the last thing it does.  This
       lets us tell when all background processes have stopped.  If
       any of these files are left from previous run, delete them */

    for(i = 1;i <= control.n_users; i++) {
         finished[i] = 1;
         results.userno = i;
         write(rfid,&results,sizeof(results));   /* for user i */
         sprintf(runfile,"%s%d",RUNFILE,i);
         if (fexists(runfile))
              if (unlink(runfile) < 0) {
                   fprintf(stderr,"iobench: Could not delete %s\n",runfile);
                   return;
                   }
         }

    close(rfid);

    if (fclose(of[1])) {
         fprintf(stderr,"iobench: Could not close %s\n",outname);
         return;
         }

    of[1] = NULL;

    /* Start up each user as a background process taking following
       arguments:

         1: the user number,
         2: the pathname of copy of control info (so it can find
            pathnames of files),
         3: the pathname of SYNCFILE
         4: the pathname of RUNFILEnn
         5: the pathname of RESULTS

       Each of the users will wait until sync file is created,
       at which time they will start the benchmark proper */

    sleep(5);
    errs = 0;
    for(i = 1; i <= control.n_users && !errs; i++) {
         userargv[0] = "iouser";
         sprintf(temp,"%d",i);
         sprintf(runfile,"%s%d",RUNFILE,i);
         userargv[1] = temp;
         userargv[2] = CTLFILE;
         userargv[3] = SYNCFILE;
         userargv[4] = runfile;
         userargv[5] = RESULTS;
         userargv[6] = NULL;

         if (pid[i] = run("iouser",userargv) > 0) {
              if (debug) printf("User %d is process %d\n",i,pid[i]);
              finished[i] = 0;
              spawned_users++;
              }
         else {

              /* Since we've been busy spawning lots of processes, assume
                 that failure to create a process means there aren't any
                 more.  Instead of quitting, continue the run with the
                 processes we've created thus far. */

              if ((of[1] = fopen(outname,"a")) == NULL) {
                   fprintf(stderr,"iobench: Could not reopen %s\n",outname);
                   errs++;
                   break;
                   }

              if ((rfid = open(RESULTS,O_RDWR)) < 0) {
                   fprintf(stderr,"iobench: Could not reopen %s\n",RESULTS);
                   errs++;
                   break;
                   }

              OFLOOP {
                   fprintf(of[z],"Could not initiate process for user %d\n",i);
                   fprintf(of[z],"Continuing run with USERS = %d\n",i-1);
                   }

              control.n_users = results.n_users = i-1;
              lseek(rfid,0,0);
              write(rfid,&results,sizeof(results));

              fclose(of[1]);
              close(rfid);
              }
         }

    if (errs) {

         /* Must kill any process we started */

         while(i > 0) killuser(i--,pid);

         return;
         }

    if (usage) {

         /* Start background process to do system monitoring passing it
            arguments specifying
             1) number of seconds sleep before starting
             2) number of seconds to gather usage statistics
             3) name of output file to be written
             4) append flag */

         monitor_system(control.start_delay,control.duration,usename,append);
         spawned_users++;
         }

    /* All the users are now running.  They will do their initialization
       and then wait for us to create the sync file before proceeding
       with benchmark proper. */

    sleep(control.start_delay);

    if (creat(SYNCFILE,0666) < 0) {
         fprintf(stderr,"iobench: Could not create %s\n",SYNCFILE);

         for (i = 1; i <= control.n_users; i++) killuser(i,pid);

         return;
         }

    printf("Released background processes for %s at %s\n",iooutput,
     cvtime(time(0),0));

    /* User processes should now be running the benchmark proper.  Wait
       for them to finish */

    sleep(control.duration + control.end_delay);

    /* Wait for all background processes to finish */

    for(loops = 0;loops < MAXLOOPS; loops++) {
         printf("Checking termination ...\n");
         n = 0;
         for(i = 1; i <= control.n_users; i++) {
              sprintf(runfile,"%s%d",RUNFILE,i);
              if (!finished[i]) {
                   if (fexists(runfile)) n++;
                   else finished[i] = 1;
                   }
              }

         if (n == 0) break;

         sleep(30);
         }

    for(i = 1; i <= control.n_users; i++)
         if (!finished[i]) {
              printf("User %d may still be running\n",i);
              killuser(i,pid);
              }

    unlink(SYNCFILE);

    if ((of[1] = fopen(outname,"a")) == NULL) {
         fprintf(stderr,"iobench: Could not reopen %s\n",outname);
         return;
         }

    hits =  statlocks(of);
    termlocks();

    server_usert = server_syst = 0.0;

    /* Flush any cached pages that need to be written to disk and
       remember how long that took */

    ft = ioflush(&flush_usert,&flush_syst);

    /* Output the statistics that user processes left in results file.
       Statistics for user K are in record K. */

    if ((rfid = open(RESULTS,O_RDONLY)) < 0) {
         fprintf(stderr,"iobench: Could not reopen %s\n",RESULTS);
         return;
         }

    for(j=0;j <= NUMBER_OF_CHECKPOINTS;j++) {
         cp = &checkpoint[j];
         cp->counter = cp->delta = 0;
         cp->usert = cp->syst = 0.0;
         }

    total_avg_resp = 0.0;
    n = total_nc = total_no = failed = maxet = 0;
    usert1 = 0.0;
    syst1 = 0.0;
    fprintf(of[1],"\n%s\n",
    " USER  #TRAN USER TIME  SYS TIME   OPENS  CLOSES  ----START TIME----   DURATION");

    for (i = 1;i <= control.n_users; i++) {
         getrec(rfid,i,&results,sizeof(results));

         if (results.counter < 0) {
              fprintf(of[1],"%5d did not write results\n",i);
              failed++;
              continue;
              }

         if (results.counter == 0) {
              fprintf(of[1],"%5d did not do any transactions\n",i);
              failed++;
              continue;
              }

         if (results.userno != i) {
              fprintf(of[1],"%5d contains data for user %d\n",i,results.userno);
              continue;
              }

         nc = no = 0;

         for(j = 0; j < control.n_files;j++) {
              opens[j] += results.opens[j];
              no += results.opens[j];
              reads[j] += results.freads[j];
              writes[j] += results.fwrites[j];

              closes[j] += results.closes[j];
              nc += results.closes[j];

              uses[j] += results.uses[j];
              sum_time[j] += results.sum_time[j];
              sumsq_time[j] += results.sumsq_time[j];

              if (results.worst_time[j] > worst_time[j]) {
                   worst_time[j] = results.worst_time[j];
                   worst_rec[j] = results.worst_rec[j];
                   worst_user[j] = i;
                   }
              }

         for(j=1;j <= NUMBER_OF_CHECKPOINTS;j++) {
              cp = &checkpoint[j];
              cp->counter += results.checkpoint[j].counter;
              cp->delta += results.checkpoint[j].delta;
              cp->usert += results.checkpoint[j].usert;
              cp->syst += results.checkpoint[j].syst;
              }

         et=results.finish-results.start;
         if (et > maxet) maxet = et;
         avg_resp= (double) et / (double) results.counter;
         fprintf(of[1],"%5d  %5d  %8.3f  %8.3f%8d%8d  %s  %8d\n",i,results.counter,
          results.usert,results.syst,no,nc,cvtime(results.start,1),et);
         n += results.counter;
         usert1 += results.usert;
         syst1 += results.syst;
         total_no += no;
         total_nc += nc;
         total_avg_resp = total_avg_resp + avg_resp;

         for (k=0;k<=25; k++) {cum_resp_dist[k] +=results.resp_dist[k];}
         }

    if (show_checkpoints) {
         fprintf(of[1],"\nCheckpoint data follows\n\n\
CP #TRANS   SECS   USER T   SYS T  Delta TPS  Aggregate TPS\n");
         for(j=1;j <= NUMBER_OF_CHECKPOINTS;j++) {
              cp = &checkpoint[j];
              cp->delta /= control.n_users;
              fprintf(of[1],"%2d %6d%7d%9.3f%8.3f",
               j,cp->counter,cp->delta,cp->usert,cp->syst);
              if (k = cp -> delta - checkpoint[j-1].delta) 
                   fprintf(of[1],"%11.3f",
                    (double)(cp->counter-checkpoint[j-1].counter)/(double)k);
              else fprintf(of[1],"   infinite");
              fprintf(of[1],"%15.3f\n",(double)(cp->counter)/(double)(cp -> delta));
              }
         }

    /* total benchmark time is sum of max actual time and flush time */

    totaltime = maxet + ft;

    if (total_nc + total_no) {
         OFLOOP fprintf(of[z],"\n%d file opens and %d file closes\n",total_no,total_nc);
         }

    OFLOOP fprintf(of[z],"\nFILE   OPENS  CLOSES    USES  AVG TIME   STD DEV     MAX T  RECORD  USER\n");
    for(j = 0;j < control.n_files; j++) {

         samples = (double)uses[j];
         avg_time = (double)sum_time[j]/samples;

         variance_time = (double)sumsq_time[j]/samples -  avg_time * avg_time;

         stddev_time = sqrt(variance_time);

         OFLOOP {
              fprintf(of[z],"%4d%8d%8d  %6d  %8.3f  ",j, opens[j],
                closes[j],uses[j],avg_time);
              fprintf(of[z],"%8.3f  %8.3f  %6d  %4d\n", stddev_time,worst_time[j],
                worst_rec[j],worst_user[j]);
              }
         }

    time(&start_time);

    mb = ((double)(n * (control.n_reads + control.n_writes) *
     control.reclen)) / 1.0E6;

    OFLOOP {
         fprintf(of[z],"\n*** Actual   time = %d seconds\n",maxet);
         fprintf(of[z],"*** Flush    time = %d seconds\n",ft);
         fprintf(of[z],"*** Adjusted time = %d seconds\n",totaltime);

         fprintf(of[z],"\n*** Total transactions = %d\n",n);
         fprintf(of[z],"*** Total MB of data = %.3f\n\n",mb);

         fprintf(of[z],"*** User       CP seconds = %.3f\n",usert1);
         fprintf(of[z],"*** System     CP seconds = %.3f\n\n",syst1);

         fprintf(of[z],"*** Server (U) CP seconds = %.3f\n",server_usert);
         fprintf(of[z],"*** Server (S) CP seconds = %.3f\n\n",server_syst);

         fprintf(of[z],"*** Flush (U)  CP seconds = %.3f\n",flush_usert);
         fprintf(of[z],"*** Flush (S)  CP seconds = %.3f\n\n",flush_syst);
         }

    usert1 += (server_usert+flush_usert);
    syst1 += (server_syst+flush_syst);

    OFLOOP {
         fprintf(of[z],"*** Total User   CP seconds = %.3f\n",usert1);
         fprintf(of[z],"*** Total System CP seconds = %.3f\n",syst1);
         fprintf(of[z],"*** Total        CP seconds = %.3f\n",usert1+syst1);
         }

    /* Show transaction statistics */

    tps = (double) n / (double) totaltime;
    total_avg_resp = total_avg_resp / (double) control.n_users;
    ut = 1000.0 * usert1 / (double) n;
    st = 1000.0 * syst1 / (double) n;
    tt = 1000.0 * (usert1+syst1) / (double) n;

    OFLOOP {
         fprintf(of[z],"\n*** Transaction statistics ***\n\n");
         fprintf(of[z],"*** %d transactions consisting of %d reads and %d writes\n",n,
          control.n_reads,control.n_writes);
         fprintf(of[z],"*** Transactions per Second = %.3f\n",tps);
         fprintf(of[z],"*** Average response time = %.3f Seconds\n",total_avg_resp);
         fprintf(of[z],"*** Average User   CP time per transaction = %.3f milliseconds\n",ut);
         fprintf(of[z],"*** Average System CP time per transaction = %.3f milliseconds\n",st);
         fprintf(of[z],"*** Average Total  CP time per transaction = %.3f milliseconds\n",tt);

         if (hits) fprintf(of[z],"*** Total number of lock hits = %d\n",hits);

         fprintf(of[z],"\n USER  AVG TIME   STD DEV  MAX TIME\n");
         }

    for(i = 1; i <= control.n_users;i++) {

         getrec(rfid,i,&results,sizeof(results));

         samples = (double)results.counter;
         avg_time = results.tran_sum_time/samples;
         variance_time = results.tran_sumsq_time/samples - avg_time * avg_time;
         stddev_time = sqrt(variance_time);

         OFLOOP fprintf(of[z],"%5d  %8.3f  %8.3f  %8.3f\n",i,avg_time,stddev_time,
          results.tran_max_time);
         }

    if (show_tps) {

         /* Show response time details */

         fprintf(of[1],"\n*** Response time distribution table\n\n");
         fprintf(of[1],"Seconds |");
         for (k=0; k<=8; k++)    fprintf (of[1]," %3d  |",k);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+-------\n");
         fprintf(of[1],"Count   |");
         for (k=0; k<=8; k++)    fprintf (of[1],"%5d |",cum_resp_dist[k]);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+-------\n");
         fprintf(of[1],"Percent |");
         for (k=0;k<=8; k++) fprintf(of[1],"%4.1f%% |",cum_resp_dist[k]/(double) n *100.0);
         fprintf(of[1],"\n\nSeconds |");
         for (k=9; k<=17; k++)    fprintf (of[1]," %3d  |",k);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+-------\n");
         fprintf(of[1],"Count   |");
         for (k=9; k<=17; k++) fprintf (of[1],"%5d |",cum_resp_dist[k]);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+-------\n");
         fprintf(of[1],"Percent |");
         for (k=9;k<=17; k++) fprintf(of[1],"%4.1f%% |",
          cum_resp_dist[k]/(double) n *100.0);
         fprintf(of[1],"\n\nSeconds |");
         for (k=18; k<=24; k++)    fprintf (of[1]," %3d  |",k);
         fprintf(of[1]," >24  |");
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+\n");
         fprintf(of[1],"Count   |");
         for (k=18; k<=25; k++) fprintf (of[1],"%5d |",cum_resp_dist[k]);
         fprintf(of[1],"\n--------+------+------+------+------+------+------+\
------+------+\n");
         fprintf(of[1],"Percent |");
         for (k=18;k<=25; k++) fprintf(of[1],"%4.1f%% |",
          cum_resp_dist[k]/(double) n *100.0);
         fprintf(of[1],"\n");
         }

    if (failed) {
         OFLOOP fprintf(of[z],"\n*** %d users did not complete normally\n",failed);

         completed = control.n_users - failed;

         avg_n = n / completed;
         avg_usert = usert1 / (double)completed;
         avg_syst = syst1 / (double)completed;

         n += failed * avg_n;
         usert1 += failed * avg_usert;
         syst1 += failed * avg_syst;

         mb = ((double)(n * 2 * control.reclen)) / 1.0E6;

         OFLOOP {
              fprintf(of[z],"\n*** Adjusted Total transactions = %d\n",n);

              fprintf(of[z],"*** Adjusted Total MB of data = %.3f\n",mb);
              fprintf(of[z],"*** Adjusted Total CP seconds = %.3f\n",usert1);
              fprintf(of[z],"*** Adjusted Total IO seconds = %.3f\n",syst1);
              }
         }

    if (iosummary) {

         /* Write line in summary file */

         old = fexists(iosummary);

         of[2] = fopen(iosummary,"a");
         if (!old) fprintf(of[2],"files\tusers\treclen\tnrecs\treads\twrites\t\
work\ttrans\tTPS\tta\ttf\tuser\tsys\tavg. rt\tfailed\thits\n");
         }

    OFLOOP {
         if (z < 2) fprintf(of[z],"\n===");

         fprintf(of[z],"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d",
          control.n_files,control.n_users,control.reclen,control.n_recs,
          control.n_reads,control.n_writes,control.param,n);
         fprintf(of[z],"\t%.3f\t%d\t%d\t%.3f\t%.3f\t%.3f\t%d\t%d\n",
          tps,maxet,ft,usert1,syst1,total_avg_resp,failed,hits);
         }

    if (iosummary) fclose(of[2]);

    unlink(RESULTS);
    unlink(CTLFILE);

    OFLOOP fprintf(of[z],"Finished %s at %s\n",iooutput,cvtime(start_time,0));
    fclose(of[1]);

    }

/* Populate a file */

populate(fn,rp)
register int fn;
register struct rec *rp;
    {
    register int curr_size,left;
    int z;
    long t1,t2;
    double usert1,usert2,syst1,syst2;

    if ((curr_size = fsize(control.path[fn])) >= control.filelen)
         return(close(unit[fn]));

    cptimes(&usert1,&syst1);
    time(&t1);

    if (curr_size) {
         OFLOOP fprintf(of[z],"Increasing size of %s from %d to %d bytes\n",control.path[fn],
          curr_size,control.filelen);
         lseek(unit[fn],0,2);               /* Seek to EOF */
         }
    else {
         OFLOOP fprintf(of[z],"Populating %s to %d bytes\n",control.path[fn],
          control.filelen);
         }

    while(curr_size+INITRECLEN <= control.filelen) {
         if (write(unit[fn],rp,INITRECLEN) < 0) {
              fprintf(stderr,"iobench: Write failed at byte %d of %s\n",
               curr_size,control.path[fn]);
              close(unit[fn]);
              return(1);
              }

         curr_size += INITRECLEN;
         }

    if (left = control.filelen - curr_size) {
         if (write(unit[fn],rp,left) < 0) {
              fprintf(stderr,"iobench: Write failed at byte %d of %s\n",
               left,control.path[fn]);
              close(unit[fn]);
              return(1);
              }
         }

    cptimes(&usert2,&syst2);
    time(&t2);
    OFLOOP fprintf(of[z],"%d seconds, %.3f user CP seconds, %.3f system CP seconds\n",t2-t1,usert2-usert1,syst2-syst1);

    return(close(unit[fn]));
    }

/* Print contents of control file */

print_control(fid)
register FILE *fid;
    {
    register int i;
    fprintf(fid,"Started %s at %s\n",iooutput,cvtime(start_time,0));
    fprintf(fid,"IOBENCH version %s built on %s\n",version,compdate);
    fprintf(fid,"IOBENCH build flags: %s\n",flags);
    if (iosummary) fprintf(fid,"Summary will be written to file %s\n",iosummary);
    fprintf(fid,"Run id is %d\n",control.runid);
    if (control.variant) fprintf(fid,"Variant is %o\n",control.variant);
    if (control.runtype == TIMED_RUN) fprintf(fid,"Timed run for %d minutes\n",control.duration);
    else fprintf(fid,"Counted run for %d transactions\nEstimated duration = %d seconds\n",
          control.trans,control.duration);
    fprintf(fid,"Number of files = %d\n",control.n_files);
    fprintf(fid,"Number of records = %d\n",control.n_recs);
    fprintf(fid,"Record length = %d\n",control.reclen);
    fprintf(fid,"File length = %d\n",control.filelen);
    fprintf(fid,"Number of users = %d\n",control.n_users);
    fprintf(fid,"Locks per file = %d\n",control.locks_per_file);
    fprintf(fid,"Lock sleep time = %d\n",control.lock_sleep_time);
    fprintf(fid,"Start delay = %d\n",control.start_delay);
    fprintf(fid,"End delay = %d\n",control.end_delay);
    fprintf(fid,"Work parameter = %d\n",control.param);
    fprintf(fid,"Reads = %d\n",control.n_reads);
    fprintf(fid,"Writes = %d\n",control.n_writes);

    if (control.calibrate)
         fprintf(fid,"Calibration run, no IO will be done\n");

    if (control.locks_per_file == 0)
         fprintf(fid,"No locking will be done\n");

    for(i=0; i < control.n_files; i++)
         fprintf(fid,"path[%d] = %s\n",i,control.path[i]);
    }

/* Dump copy of control information into specified file.  Return 1 if error,
   0 otherwise */

dump_control(fname)
char *fname;
    {
    register FILE *fid;
    register int i;

    unlink(fname);

    if ((fid = fopen(fname,"w")) == NULL) return(1);

    fprintf(fid,"%d\n",control.runid);
    fprintf(fid,"%d\n",control.runtype);
    fprintf(fid,"%d\n",control.n_files);
    fprintf(fid,"%d\n",control.n_recs);
    fprintf(fid,"%d\n",control.reclen);
    fprintf(fid,"%d\n",control.filelen);
    fprintf(fid,"%d\n",control.n_users);
    fprintf(fid,"%d\n",control.locks_per_file);
    fprintf(fid,"%d\n",control.lock_sleep_time);
    fprintf(fid,"%d\n",control.start_delay);
    fprintf(fid,"%d\n",control.duration);
    fprintf(fid,"%d\n",control.end_delay);
    fprintf(fid,"%d\n",control.param);
    fprintf(fid,"%d\n",control.n_reads);
    fprintf(fid,"%d\n",control.n_writes);
    fprintf(fid,"%d\n",control.trans);
    fprintf(fid,"%d\n",control.variant);

    for(i=0; i < control.n_files; i++)
         fprintf(fid,"%s\n",control.path[i]);

    return(fclose(fid) < 0 ? 1 : 0);
    }

/* Kill specified user */

killuser(un,pid)
int un,*pid;
    {
    if (kill(pid[un],SIGTERM) == 0) fprintf(stderr,"Killed user %d, process %d\n",
          un,pid[un]);
    else fprintf(stderr,"Could not kill user %d, process %d, errno = %d\n",
          un,pid[un],errno);
    }

int cleanup()
    {
    register int i;

    termlocks();

    exit(0);
    }
